import asyncio
import logging
import re
import threading
import time
import traceback
from collections import deque
from collections.abc import Awaitable, Callable, Coroutine
from contextlib import asynccontextmanager
from datetime import UTC, datetime
from functools import wraps
from typing import Any, ParamSpec, TypeVar, cast

# Define generic type variables for return type and parameters
R = TypeVar('R')
P = ParamSpec('P')
QueueEntryType = TypeVar('QueueEntryType')

_MONOTONIC_DATETIME_REGEX = re.compile(
    r'^(\d{4})-(\d{2})-(\d{2})'
    r'T(\d{2}):(\d{2}):(\d{2})'
    r'(?:\.(\d{1,9}))?'
    r'(Z|[+-]\d{2}:\d{2})$'
)
_MONOTONIC_DATETIME_LENGTH = 30  # YYYY-MM-DDTHH:MM:SS.fffffffffZ
_MONOTONIC_EPOCH_ANCHOR_NS = time.time_ns()
_MONOTONIC_CLOCK_ANCHOR_NS = time.monotonic_ns()
_last_monotonic_datetime_ns = _MONOTONIC_EPOCH_ANCHOR_NS
_last_monotonic_datetime_lock = threading.Lock()


def _format_epoch_ns_to_iso(epoch_ns: int) -> str:
    seconds, fractional_ns = divmod(epoch_ns, 1_000_000_000)
    base_datetime = datetime.fromtimestamp(seconds, tz=UTC)
    if base_datetime.year <= 1990 or base_datetime.year >= 2500:
        raise ValueError('Datetime year must be >1990 and <2500')
    normalized = f'{base_datetime.strftime("%Y-%m-%dT%H:%M:%S")}.{fractional_ns:09d}Z'
    assert len(normalized) == _MONOTONIC_DATETIME_LENGTH, (
        f'Expected canonical datetime length {_MONOTONIC_DATETIME_LENGTH}, got {len(normalized)}: {normalized!r}'
    )
    return normalized


def monotonic_datetime(isostring: str | None = None) -> str:
    """Return canonical UTC ISO datetime with exactly 9 fractional digits."""
    if isostring is not None:
        match = _MONOTONIC_DATETIME_REGEX.fullmatch(isostring)
        if match is None:
            raise ValueError(f'Invalid ISO datetime: {isostring!r}')

        parsed = datetime.fromisoformat(isostring.replace('Z', '+00:00'))
        parsed_utc = parsed.astimezone(UTC)
        if parsed_utc.year <= 1990 or parsed_utc.year >= 2500:
            raise ValueError(f'Datetime year must be >1990 and <2500: {isostring!r}')

        fractional = (match.group(7) or '').ljust(9, '0')
        normalized = f'{parsed_utc.strftime("%Y-%m-%dT%H:%M:%S")}.{fractional}Z'
        assert len(normalized) == _MONOTONIC_DATETIME_LENGTH, (
            f'Expected canonical datetime length {_MONOTONIC_DATETIME_LENGTH}, got {len(normalized)}: {normalized!r}'
        )
        return normalized

    global _last_monotonic_datetime_ns
    elapsed_ns = time.monotonic_ns() - _MONOTONIC_CLOCK_ANCHOR_NS
    epoch_ns = _MONOTONIC_EPOCH_ANCHOR_NS + elapsed_ns
    with _last_monotonic_datetime_lock:
        if epoch_ns <= _last_monotonic_datetime_ns:
            epoch_ns = _last_monotonic_datetime_ns + 1
        _last_monotonic_datetime_ns = epoch_ns
    return _format_epoch_ns_to_iso(epoch_ns)


async def run_with_timeout(awaitable: Awaitable[R], timeout: float | None = None) -> R:
    """Await `awaitable` with optional timeout."""
    if timeout is None:
        return await awaitable
    return await asyncio.wait_for(awaitable, timeout=timeout)


async def cancel_and_await(task: asyncio.Task[Any] | None, timeout: float | None = None) -> None:
    """Best-effort task cancellation helper that suppresses cancellation-time noise."""
    if task is None:
        return
    if not task.done():
        task.cancel()
    try:
        await run_with_timeout(task, timeout=timeout)
    except (asyncio.CancelledError, TimeoutError):
        pass
    except Exception:
        pass


@asynccontextmanager
async def _run_with_slow_monitor(
    monitor_factory: Callable[[], Coroutine[Any, Any, Any]] | None,
    *,
    task_name: str | None = None,
):
    """Run an optional slow-monitor task scoped to the surrounding execution.

    The monitor is started on enter and always cancelled on exit.
    """
    task: asyncio.Task[Any] | None = None
    if monitor_factory is not None:
        if task_name is None:
            task = asyncio.create_task(monitor_factory())
        else:
            task = asyncio.create_task(monitor_factory(), name=task_name)
    try:
        yield task
    finally:
        await cancel_and_await(task)


class QueueShutDown(Exception):
    """Raised when putting on to or getting from a shut-down Queue."""

    pass


class CleanShutdownQueue(asyncio.Queue[QueueEntryType]):
    """asyncio.Queue subclass that handles shutdown cleanly without warnings."""

    _is_shutdown: bool = False
    _queue: deque[QueueEntryType]
    _getters: deque[asyncio.Future[QueueEntryType]]
    _putters: deque[asyncio.Future[QueueEntryType]]

    def iter_items(self) -> tuple[QueueEntryType, ...]:
        """Return a snapshot of queued items in FIFO order."""
        return tuple(self._queue)

    def remove_item(self, item: QueueEntryType) -> bool:
        """Remove one matching queued item if present."""
        if item not in self._queue:
            return False
        self._queue.remove(item)
        return True

    def shutdown(self, immediate: bool = True):
        """Shutdown the queue and clean up all pending futures."""
        del immediate
        self._is_shutdown = True

        # Cancel all waiting getters without triggering warnings
        while self._getters:
            getter = self._getters.popleft()
            if not getter.done():
                # Set exception instead of cancelling to avoid "Event loop is closed" errors
                getter.set_exception(QueueShutDown())

        # Cancel all waiting putters
        while self._putters:
            putter = self._putters.popleft()
            if not putter.done():
                putter.set_exception(QueueShutDown())

    async def get(self) -> QueueEntryType:
        """Remove and return an item from the queue, with shutdown support."""
        while self.empty():
            if self._is_shutdown:
                raise QueueShutDown

            getter: asyncio.Future[QueueEntryType] = asyncio.get_running_loop().create_future()
            assert isinstance(getter, asyncio.Future)
            self._getters.append(getter)
            try:
                await getter
            except:
                # Clean up the getter if we're cancelled
                getter.cancel()  # Just in case getter is not done yet.
                try:
                    self._getters.remove(getter)
                except ValueError:
                    pass
                # Re-raise the exception
                raise

        return self.get_nowait()

    async def put(self, item: QueueEntryType) -> None:
        """Put an item into the queue, with shutdown support."""
        while self.full():
            if self._is_shutdown:
                raise QueueShutDown

            putter: asyncio.Future[QueueEntryType] = asyncio.get_running_loop().create_future()
            assert isinstance(putter, asyncio.Future)
            self._putters.append(putter)
            try:
                await putter
            except:
                putter.cancel()  # Just in case putter is not done yet.
                try:
                    self._putters.remove(putter)
                except ValueError:
                    pass
                raise

        return self.put_nowait(item)

    def put_nowait(self, item: QueueEntryType) -> None:
        """Put an item into the queue without blocking, with shutdown support."""
        if self._is_shutdown:
            raise QueueShutDown
        return super().put_nowait(item)

    def get_nowait(self) -> QueueEntryType:
        """Remove and return an item if one is immediately available, with shutdown support."""
        if self._is_shutdown and self.empty():
            raise QueueShutDown
        return super().get_nowait()


def extract_basemodel_generic_arg(cls: type) -> Any:
    """
    Extract T_EventResultType Generic arg from BaseEvent[T_EventResultType] subclasses using pydantic generic metadata.
    Needed because pydantic messes with the mro and obscures the Generic from the bases list.
    https://github.com/pydantic/pydantic/issues/8410
    """

    def _extract_arg_from_metadata(metadata_value: Any) -> Any:
        metadata = cast(dict[str, Any], metadata_value)
        origin: Any = metadata.get('origin')
        args: tuple[Any, ...] = cast(tuple[Any, ...], metadata.get('args') or ())
        if not args:
            return None
        # Avoid importing BaseEvent here to keep helpers.py decoupled from models.py.
        if getattr(origin, '__name__', None) == 'BaseEvent' and getattr(origin, '__module__', None) == 'abxbus.base_event':
            return args[0]
        return None

    # Direct check first for speed - most subclasses will have it directly
    if hasattr(cls, '__pydantic_generic_metadata__'):
        generic_arg = _extract_arg_from_metadata(getattr(cls, '__pydantic_generic_metadata__'))
        if generic_arg is not None:
            return generic_arg

    # Only check MRO if direct check failed
    for parent in cls.__mro__[1:]:
        if hasattr(parent, '__pydantic_generic_metadata__'):
            generic_arg = _extract_arg_from_metadata(getattr(parent, '__pydantic_generic_metadata__'))
            if generic_arg is not None:
                return generic_arg

    return None


def time_execution(
    additional_text: str = '',
) -> Callable[[Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]]:
    """Decorator that logs how much time execution of a function takes"""

    def decorator(func: Callable[P, Coroutine[Any, Any, R]]) -> Callable[P, Coroutine[Any, Any, R]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            start_time = time.time()
            result = await func(*args, **kwargs)
            execution_time = time.time() - start_time
            # Only log if execution takes more than 0.25 seconds to avoid spamming the logs
            # you can lower this threshold locally when you're doing dev work to performance optimize stuff
            if execution_time > 0.25:
                self_has_logger = args and getattr(args[0], 'logger', None)
                if self_has_logger:
                    logger = getattr(args[0], 'logger')
                elif 'agent' in kwargs:
                    logger = getattr(kwargs['agent'], 'logger')
                elif 'browser_session' in kwargs:
                    logger = getattr(kwargs['browser_session'], 'logger')
                else:
                    logger = logging.getLogger(__name__)
                logger.debug(f'⏳ {additional_text.strip("-")}() took {execution_time:.2f}s')
            return result

        return wrapper

    return decorator


def log_filtered_traceback(exc: BaseException) -> str:
    """Format traceback while filtering noisy asyncio/stdlib frames."""
    trace_exc = traceback.TracebackException.from_exception(exc, capture_locals=False)

    def _filter(_: traceback.TracebackException):
        trace_exc.stack = traceback.StackSummary.from_list(
            [f for f in trace_exc.stack if 'asyncio/tasks.py' not in f.filename and 'lib/python' not in f.filename]
        )
        if trace_exc.__cause__:
            _filter(trace_exc.__cause__)
        if trace_exc.__context__:
            _filter(trace_exc.__context__)

    _filter(trace_exc)
    return ''.join(trace_exc.format())


__all__ = [
    'run_with_timeout',
    'cancel_and_await',
    '_run_with_slow_monitor',
    'log_filtered_traceback',
    'CleanShutdownQueue',
    'QueueShutDown',
    'extract_basemodel_generic_arg',
    'time_execution',
]



"""Helper functions for logging event trees and formatting"""

import asyncio
import logging
import math
from collections import defaultdict
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from abxbus.base_event import BaseEvent, EventResult
    from abxbus.event_bus import EventBus


def format_timestamp(dt: str | datetime | None) -> str:
    """Format an ISO datetime string (or datetime) for display."""
    if dt is None:
        return 'N/A'
    if isinstance(dt, str):
        parsed = datetime.fromisoformat(dt)
    else:
        parsed = dt
    return parsed.strftime('%H:%M:%S.%f')[:-3]


def format_result_value(value: Any) -> str:
    """Format a result value for display"""
    if value is None:
        return 'None'
    if hasattr(value, 'event_type') and hasattr(value, 'event_id'):  # BaseEvent check without import
        return f'Event({value.event_type}#{value.event_id[-4:]})'
    if isinstance(value, (str, int, float, bool)):
        return repr(value)
    if isinstance(value, dict):
        value_dict = cast(dict[Any, Any], value)
        return f'dict({len(value_dict)} items)'
    if isinstance(value, list):
        value_list = cast(list[Any], value)
        return f'list({len(value_list)} items)'
    return f'{type(value).__name__}(...)'


def log_event_tree(
    event: 'BaseEvent[Any]',
    indent: str = '',
    is_last: bool = True,
    event_children_by_parent: dict[str | None, list['BaseEvent[Any]']] | None = None,
) -> str:
    from abxbus.base_event import logger

    """Print this event and its results with proper tree formatting"""
    # Determine the connector
    connector = '└── ' if is_last else '├── '

    # Print this event's line
    # Format timing info
    timing_str = f'[{format_timestamp(event.event_created_at)}'
    if event.event_completed_at and event.event_created_at:
        completed_dt = datetime.fromisoformat(event.event_completed_at)
        created_dt = datetime.fromisoformat(event.event_created_at)
        duration = (completed_dt - created_dt).total_seconds()
        timing_str += f' ({duration:.3f}s)'
    timing_str += ']'

    lines: list[str] = []

    event_line = f'{indent}{connector}{event.event_type}#{event.event_id[-4:]} {timing_str}'
    logger.warning(event_line)
    lines.append(event_line)

    # Calculate the new indent for children
    extension = '    ' if is_last else '│   '
    new_indent = indent + extension

    # Track which child events were printed via handlers to avoid duplicates
    printed_child_ids: set[str] = set()

    # Print each result
    if event.event_results:
        results_sorted = sorted(event.event_results.items(), key=lambda x: x[1].started_at or '')

        # Calculate which is the last item considering both results and unmapped children
        unmapped_children: list[BaseEvent[Any]] = []
        if event_children_by_parent:
            all_children = event_children_by_parent.get(event.event_id, [])
            for child in all_children:
                # Will be printed later if not already printed by a handler
                if child.event_id not in [c.event_id for r in event.event_results.values() for c in r.event_children]:
                    unmapped_children.append(child)

        total_items = len(results_sorted) + len(unmapped_children)

        for i, (_handler_id, result) in enumerate(results_sorted):
            is_last_item = i == total_items - 1
            lines.append(log_event_result_tree(result, new_indent, is_last_item, event_children_by_parent))
            # Track child events printed by this result
            for child in result.event_children:
                printed_child_ids.add(child.event_id)

    # Print unmapped children (those not printed by any handler)
    if event_children_by_parent:
        children = event_children_by_parent.get(event.event_id, [])
        for i, child in enumerate(children):
            if child.event_id not in printed_child_ids:
                is_last_child = i == len(children) - 1
                lines.append(log_event_tree(child, new_indent, is_last_child, event_children_by_parent))

    return '\n'.join(lines)


def log_event_result_tree(
    result: 'EventResult[Any]',
    indent: str = '',
    is_last: bool = True,
    event_children_by_parent: dict[str | None, list['BaseEvent[Any]']] | None = None,
) -> str:
    """Print this result and its child events with proper tree formatting"""

    from abxbus.base_event import logger

    # Determine the connector
    connector = '└── ' if is_last else '├── '

    # Status icon
    result_icon = (
        '✅'
        if result.status == 'completed'
        else '❌'
        if result.error is not None
        else '🏃'
        if result.status == 'started'
        else '⏳'
    )

    # Format handler name with bus info
    handler_display = f'{result.eventbus_label}.{result.handler.label}'

    # Format the result line
    result_line = f'{indent}{connector}{result_icon} {handler_display}'

    # Add timing info
    if result.started_at:
        result_line += f' [{format_timestamp(result.started_at)}'
        if result.completed_at:
            completed_dt = datetime.fromisoformat(result.completed_at)
            started_dt = datetime.fromisoformat(result.started_at)
            duration = (completed_dt - started_dt).total_seconds()
            result_line += f' ({duration:.3f}s)'
        result_line += ']'

    # Add result value or error
    if result.status == 'error' and result.error:
        result_line += f' ☠️ {type(result.error).__name__}: {str(result.error)}'
    elif result.status == 'completed':
        result_line += f' → {format_result_value(result.result)}'

    lines: list[str] = []
    logger.warning(result_line)
    lines.append(result_line)

    # Calculate the new indent for child events
    extension = '    ' if is_last else '│   '
    new_indent = indent + extension

    # Print child events dispatched by this handler

    if result.event_children:
        for i, child in enumerate(result.event_children):
            is_last_child = i == len(result.event_children) - 1
            lines.append(log_event_tree(child, new_indent, is_last_child, event_children_by_parent))

    return '\n'.join(lines)


def log_eventbus_tree(eventbus: 'EventBus') -> str:
    """Print a nice pretty formatted tree view of all events in the history including their results and child events recursively"""

    from abxbus.base_event import logger

    # Build a mapping of parent_id to child events
    parent_to_children: dict[str | None, list[BaseEvent[Any]]] = defaultdict(list)
    for event in eventbus.event_history.values():
        parent_to_children[event.event_parent_id].append(event)

    # Sort events by creation time
    for children in parent_to_children.values():
        children.sort(key=lambda e: e.event_created_at)

    # Find root events (those without parents or with self as parent)
    root_events = list(parent_to_children[None])

    # Also include events that have themselves as parent (edge case)
    for event in eventbus.event_history.values():
        if event.event_parent_id == event.event_id and event not in root_events:
            root_events.append(event)
            # Remove from its incorrect parent mapping to avoid double printing
            if event.event_id in parent_to_children:
                parent_to_children[event.event_id] = [
                    e for e in parent_to_children[event.event_id] if e.event_id != event.event_id
                ]

    logger.warning(f'📊 Event History Tree for {eventbus}')
    logger.warning('=' * 80)

    if not root_events:
        logger.warning('  (No events in history)')
        return '(No events in history)'

    # Print all root events using their log_tree helper method
    lines: list[str] = []
    for i, event in enumerate(root_events):
        is_last = i == len(root_events) - 1
        lines.append(log_event_tree(event, '', is_last, parent_to_children))

    logger.warning('=' * 80)

    return '\n'.join(lines)


def log_timeout_tree(event: 'BaseEvent[Any]', timed_out_result: 'EventResult[Any]') -> None:
    """Log detailed timeout information showing the event tree and which handler timed out"""

    from abxbus.base_event import logger
    from abxbus.event_bus import EventBus

    if not logger.isEnabledFor(logging.WARNING):
        return

    now = datetime.now(UTC)

    # Find the root event by walking up the parent chain
    root_event = event
    visited_parent_ids: set[str] = set()
    while root_event.event_parent_id and root_event.event_parent_id not in visited_parent_ids:
        visited_parent_ids.add(root_event.event_parent_id)
        parent_found = False
        # Search for parent in all EventBus instances
        for bus in list(EventBus.all_instances):
            if root_event.event_parent_id in bus.event_history:
                root_event = bus.event_history[root_event.event_parent_id]
                parent_found = True
                break
        if not parent_found:
            break

    red = '\033[91m'
    green = '\033[92m'
    yellow = '\033[93m'
    pink = '\033[95m'
    reset = '\033[0m'

    logger.warning('=' * 80)
    logger.warning(
        f'⏱️  TIMEOUT ERROR - Handling took more than {event.event_timeout}s for {timed_out_result.eventbus_label}.{timed_out_result.handler_name}({event})'
    )
    logger.warning('=' * 80)

    def print_handler_line(
        handler_indent: str,
        handler_name: str,
        event_id_suffix: str,
        status: str = 'pending',
        started_at: str | None = None,
        completed_at: str | None = None,
        timeout: float | None = None,
        is_expired: bool = False,
        is_interrupted: bool = False,
        is_pending: bool = False,
        error_type: str | None = None,
    ):
        """Print a formatted handler line with proper column alignment"""

        # Col 2: icon based on status
        if status == 'completed':
            col2_icon = '☑️'
        elif is_pending:
            col2_icon = '🔚'
        elif status == 'started' or is_interrupted:
            col2_icon = '➡️'
        elif is_expired:
            col2_icon = '⏰'
        elif status == 'error':
            col2_icon = '❌'
        else:
            col2_icon = '🔜'

        # Col 3: handler description
        col3_desc = f'{handler_name}(#{event_id_suffix})'

        # Col 4: padding to column 64
        left_part = f'{handler_indent}{col2_icon} {col3_desc}'
        col4_padding = ' ' * max(1, 64 - len(f'{handler_indent}   {col3_desc}'))  # assume icons are always 2 chars wide

        # Col 5-10: timing info
        max_time = timeout or 0
        if started_at:
            started_at_dt = datetime.fromisoformat(started_at)
            completed_at_dt = datetime.fromisoformat(completed_at) if completed_at else now
            elapsed_time = (completed_at_dt - started_at_dt).total_seconds()

            if is_expired or (elapsed_time >= max_time):
                col5_timing_icon = '⌛️'
                if is_expired:
                    col9_extra = f' ⬅️ {red}TIMEOUT HERE{reset} ⏰'
                else:
                    col9_extra = f' ☠️ {pink}{error_type or "FAILED"}{reset}'  # timed out before us, but unrelated to current timeout exception chain, not the direct cause of our current error
            elif is_interrupted and is_pending:
                col5_timing_icon = '  '
                col9_extra = f' ⛔️ {pink}{error_type or "AbortedError"}{reset}'
            elif is_interrupted:
                col5_timing_icon = '⏳'
                col9_extra = f' ⬅️ {yellow}INTERRUPTED{reset} ✂️'
            elif status == 'started':
                col5_timing_icon = '⏳'
                col9_extra = ''
            else:
                col5_timing_icon = '  '
                col9_extra = ' ✓' if status == 'completed' else '    ✗'

            if elapsed_time >= max_time and not is_pending:
                col6_elapsed = f'{red}{round(elapsed_time):2d}s{reset}'
            elif elapsed_time > 3:
                col6_elapsed = f'{yellow}{round(elapsed_time):2d}s{reset}'
            elif status == 'completed':
                col6_elapsed = f'{green}{round(elapsed_time):2d}s{reset}'
            elif is_pending:
                col6_elapsed = '   '
            elif is_interrupted or (status == 'error' and elapsed_time <= max_time):
                col6_elapsed = f'{yellow}{round(elapsed_time):2d}s{reset}'
            else:
                col6_elapsed = f'{round(elapsed_time):2d}s'

            col7_slash = '/'
            if is_expired or elapsed_time >= max_time:
                col8_max = f'{red}{math.ceil(timeout or 0):2d}s{reset}'
            else:
                col8_max = f'{math.ceil(timeout or 0):2d}s'
        else:
            # Never started - pending
            col5_timing_icon = '🔜'
            col6_elapsed = '   '
            col7_slash = '/'
            col8_max = f'{math.ceil(timeout or 0):2d}s'
            col9_extra = ''

        # Assemble and print
        logger.warning(f'{left_part}{col4_padding}{col5_timing_icon} {col6_elapsed}{col7_slash}{col8_max}  {col9_extra}')

    def print_event_tree(evt: 'BaseEvent[Any]', indent: str = ''):
        """Recursively print event and its handlers"""
        event_start_time = (
            min(
                (result.started_at for result in evt.event_results.values() if result.started_at is not None),
                default=evt.event_created_at,
            )
            or evt.event_created_at
        )
        now = datetime.now(UTC)
        event_start_time_dt = datetime.fromisoformat(event_start_time)
        elapsed = round((now - event_start_time_dt).total_seconds())

        # Event line formatted with proper columns
        # Col 1: indent, Col 2: icon (📣), Col 3: description
        col1_indent = indent
        col2_icon = '📣'
        col3_desc = f'{evt.event_type}#{evt.event_id[-4:]}'

        # Col 4: padding to column 70
        left_part = f'{col1_indent}{col2_icon} {col3_desc}'
        col4_padding = ' ' * max(1, 64 - len(f'{col1_indent}   {col3_desc}'))

        # Col 5-9: timing info
        col5_timing_icon = '   '  # No icon for event lines
        if elapsed >= ((evt.event_timeout or 1) * len(evt.event_results)):
            col6_elapsed = f'{red}{elapsed:2d}s{reset}'
        elif elapsed > 5 or evt.event_status == 'started':
            col6_elapsed = f'{yellow}{elapsed:2d}s{reset}'
        elif evt.event_status == 'completed' and all(result.error is None for result in evt.event_results.values()):
            col6_elapsed = f'{green}{elapsed:2d}s{reset}'
        else:
            col6_elapsed = f'{elapsed:2d}s'

        # col7_slash = '/'
        # col8_max = f'{round(evt.event_timeout or 1) * len(evt.event_results):2d}s'
        # if evt.event_status == 'completed':
        #     col8_max = f'{green}{col8_max}{reset}'
        # elif evt.event_status == 'error':
        #     col8_max = f'{red}{col8_max}{reset}'
        # elif evt.event_status == 'started':
        #     col8_max = f'{yellow}{col8_max}{reset}'
        # else:
        #     col8_max = f'{col8_max}{reset}'

        # Assemble and print
        logger.warning(f'{left_part}{col4_padding}{col5_timing_icon}    {col6_elapsed}')

        # Increase indent for handlers (3 spaces to align under event name)
        handler_indent = indent + '   '

        # Get all handlers for this event
        for result in evt.event_results.values():
            # Check if this is the exact handler that timed out
            is_expired = result.handler_id == timed_out_result.handler_id

            # Check if this handler was interrupted (started but not completed, in a child of the timed-out handler)
            is_interrupted = False
            if result.status == 'error' and isinstance(result.error, asyncio.CancelledError):
                # Check if this result is in a child event of the timed-out handler
                for timed_out_child in timed_out_result.event_children:
                    if evt.event_id == timed_out_child.event_id:
                        is_interrupted = True
                        break

            # Print the handler line using helper function
            print_handler_line(
                handler_indent=handler_indent,
                handler_name=result.handler_name,
                event_id_suffix=result.event_id[-4:],
                status=result.status,
                started_at=result.started_at,
                completed_at=result.completed_at,
                timeout=result.timeout or evt.event_timeout,
                is_expired=is_expired,
                is_interrupted=is_interrupted,
                is_pending='pending' in str(result.error),
                error_type=type(result.error).__name__ if result.error else None,
            )

            # Print child events dispatched by this handler
            for child_event in result.event_children:
                print_event_tree(child_event, handler_indent + '   ')

        # After showing all handlers that ran, show any registered handlers that never started
        # This is for handlers that were registered but didn't get to run due to timeouts
        # Find which EventBus contains this event
        event_bus = None
        for bus in list(EventBus.all_instances):
            if evt.event_id in bus.event_history:
                event_bus = bus
                break

        # Get all registered handlers that could match this event_type.
        if event_bus is not None:
            indexed_ids = list(event_bus.handlers_by_key.get(evt.event_type, [])) + list(event_bus.handlers_by_key.get('*', []))

            for handler_id in indexed_ids:
                entry = event_bus.handlers.get(handler_id)
                if entry is None:
                    continue
                # Check if this handler already ran (has an EventResult)
                if handler_id not in evt.event_results:
                    # This handler was registered but never started - use helper to format
                    print_handler_line(
                        handler_indent=handler_indent,
                        handler_name=entry.handler_name,
                        event_id_suffix=evt.event_id[-4:],
                        status='pending',  # Will show 🔲 icon
                        started_at=None,
                        completed_at=None,
                        timeout=evt.event_timeout,
                        is_expired=False,
                        is_interrupted=False,
                    )

    # Print the tree starting from root
    print_event_tree(root_event)

    logger.warning('\n' + '=' * 80 + '\n')



"""IPC bridges for forwarding EventBus instances over HTTP or unix sockets."""

from __future__ import annotations

import asyncio
import importlib
import json
import logging
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal
from urllib.parse import urlparse
from urllib.request import Request, urlopen

from anyio import Path as AnyPath
from uuid_extensions import uuid7str

from abxbus.base_event import BaseEvent
from abxbus.bridge_jsonl import JSONLEventBridge
from abxbus.bridge_sqlite import SQLiteEventBridge
from abxbus.event_bus import EventBus, EventPatternType, in_handler_context

logger = logging.getLogger('abxbus.bridges')
UNIX_SOCKET_MAX_PATH_CHARS = 90
__all__ = [
    'EventBridge',
    'HTTPEventBridge',
    'SocketEventBridge',
    'JSONLEventBridge',
    'SQLiteEventBridge',
    'NATSEventBridge',
    'RedisEventBridge',
    'PostgresEventBridge',
]

if TYPE_CHECKING:
    from .bridge_nats import NATSEventBridge
    from .bridge_postgres import PostgresEventBridge
    from .bridge_redis import RedisEventBridge

_LAZY_BRIDGE_MODULES: dict[str, str] = {
    'NATSEventBridge': '.bridge_nats',
    'RedisEventBridge': '.bridge_redis',
    'PostgresEventBridge': '.bridge_postgres',
}


class _Endpoint:
    def __init__(
        self,
        raw: str,
        scheme: Literal['unix', 'http', 'https'],
        *,
        host: str | None = None,
        port: int | None = None,
        path: str | None = None,
    ):
        self.raw = raw
        self.scheme = scheme
        self.host = host
        self.port = port
        self.path = path


def _parse_endpoint(raw_endpoint: str) -> _Endpoint:
    parsed = urlparse(raw_endpoint)
    scheme = parsed.scheme.lower()

    if scheme == 'unix':
        socket_path = parsed.path or parsed.netloc
        if not socket_path:
            raise ValueError(f'Invalid unix endpoint (missing socket path): {raw_endpoint}')
        socket_path_len = len(socket_path.encode('utf-8'))
        if socket_path_len > UNIX_SOCKET_MAX_PATH_CHARS:
            raise ValueError(
                f'Unix socket path is too long ({socket_path_len} chars), max is {UNIX_SOCKET_MAX_PATH_CHARS}: {socket_path}'
            )
        return _Endpoint(raw_endpoint, 'unix', path=socket_path)

    if scheme == 'http':
        if not parsed.hostname:
            raise ValueError(f'Invalid HTTP endpoint (missing hostname): {raw_endpoint}')
        request_path = parsed.path or '/'
        if parsed.query:
            request_path = f'{request_path}?{parsed.query}'
        port = parsed.port if parsed.port is not None else 80
        return _Endpoint(raw_endpoint, 'http', host=parsed.hostname, port=port, path=request_path)

    if scheme == 'https':
        if not parsed.hostname:
            raise ValueError(f'Invalid HTTP endpoint (missing hostname): {raw_endpoint}')
        request_path = parsed.path or '/'
        if parsed.query:
            request_path = f'{request_path}?{parsed.query}'
        port = parsed.port if parsed.port is not None else 443
        return _Endpoint(raw_endpoint, 'https', host=parsed.hostname, port=port, path=request_path)

    raise ValueError(f'Unsupported endpoint scheme: {raw_endpoint}')


class EventBridge:
    """Shared bridge implementation exposing EventBus-like on/emit/dispatch."""

    def __init__(
        self,
        send_to: str | None = None,
        listen_on: str | None = None,
        *,
        name: str | None = None,
    ):
        self.send_to = _parse_endpoint(send_to) if send_to else None
        self.listen_on = _parse_endpoint(listen_on) if listen_on else None
        internal_name = name or f'EventBridge_{uuid7str()[-8:]}'
        self._inbound_bus = EventBus(name=internal_name, max_history_size=0)

        self._server: asyncio.AbstractServer | None = None
        self._start_lock = asyncio.Lock()
        self._listen_socket_path: Path | None = None
        self._autostart_task: asyncio.Task[None] | None = None

    def on(self, event_pattern: EventPatternType, handler: Callable[[BaseEvent[Any]], Any]) -> None:
        self._ensure_listener_started()
        self._inbound_bus.on(event_pattern, handler)

    async def emit(self, event: BaseEvent[Any]) -> BaseEvent[Any] | None:
        if self.send_to is None:
            raise RuntimeError(f'{self.__class__.__name__}.emit() requires send_to=...')

        payload = event.model_dump(mode='json')

        if self.send_to.scheme == 'unix':
            await self._send_unix(self.send_to, payload)
        else:
            await self._send_http(self.send_to, payload)

        if in_handler_context():
            return None
        return event

    async def dispatch(self, event: BaseEvent[Any]) -> BaseEvent[Any] | None:
        return await self.emit(event)

    async def start(self) -> None:
        if self.listen_on is None or self._server is not None:
            return

        async with self._start_lock:
            if self._server is not None:
                return

            endpoint = self.listen_on
            assert endpoint is not None
            if endpoint.scheme == 'unix':
                socket_path = Path(endpoint.path or '')
                if not socket_path.is_absolute():
                    raise ValueError(f'unix listen_on path must be absolute, got: {endpoint.raw}')
                socket_path.parent.mkdir(parents=True, exist_ok=True)
                async_socket_path = AnyPath(socket_path)
                if await async_socket_path.exists():
                    await async_socket_path.unlink()
                self._listen_socket_path = socket_path
                self._server = await asyncio.start_unix_server(self._handle_unix_client, path=str(socket_path))
                return

            if endpoint.scheme != 'http':
                raise ValueError(f'listen_on only supports unix:// or http:// endpoints, got: {endpoint.raw}')
            assert endpoint.host is not None
            assert endpoint.port is not None
            self._server = await asyncio.start_server(self._handle_http_client, host=endpoint.host, port=endpoint.port)

    async def close(self, *, clear: bool = True) -> None:
        if self._autostart_task is not None:
            await asyncio.gather(self._autostart_task, return_exceptions=True)
            self._autostart_task = None

        if self._server is not None:
            self._server.close()
            await self._server.wait_closed()
            self._server = None

        if self._listen_socket_path and self._listen_socket_path.exists():
            self._listen_socket_path.unlink()
            self._listen_socket_path = None

        await self._inbound_bus.stop(clear=clear)

    def _ensure_listener_started(self) -> None:
        if self.listen_on is None or self._server is not None:
            return
        if self._autostart_task is not None and not self._autostart_task.done():
            return
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            return
        self._autostart_task = loop.create_task(self.start())

    async def _handle_unix_client(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        try:
            while True:
                line = await reader.readline()
                if not line:
                    break
                if not line.strip():
                    continue
                await self._handle_incoming_bytes(line)
        finally:
            writer.close()
            await writer.wait_closed()

    async def _handle_http_client(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        try:
            raw_headers = await reader.readuntil(b'\r\n\r\n')
        except asyncio.IncompleteReadError:
            await self._write_http_response(writer, status=400, body='incomplete request')
            return
        except asyncio.LimitOverrunError:
            await self._write_http_response(writer, status=400, body='headers too large')
            return

        header_lines = raw_headers.decode('utf-8', errors='replace').split('\r\n')
        if not header_lines or not header_lines[0]:
            await self._write_http_response(writer, status=400, body='missing request line')
            return

        parts = header_lines[0].split(' ', 2)
        if len(parts) != 3:
            await self._write_http_response(writer, status=400, body='invalid request line')
            return
        method, request_target, _version = parts

        headers: dict[str, str] = {}
        for line in header_lines[1:]:
            if not line:
                continue
            name, separator, value = line.partition(':')
            if not separator:
                continue
            headers[name.strip().lower()] = value.strip()

        if method.upper() != 'POST':
            await self._write_http_response(writer, status=405, body='method not allowed')
            return

        expected_path = (self.listen_on.path if self.listen_on else None) or '/'
        if request_target != expected_path:
            await self._write_http_response(writer, status=404, body='not found')
            return

        content_length = headers.get('content-length', '0')
        try:
            body_size = int(content_length)
        except ValueError:
            await self._write_http_response(writer, status=400, body='invalid content-length')
            return

        if body_size < 0:
            await self._write_http_response(writer, status=400, body='invalid content-length')
            return

        try:
            body = await reader.readexactly(body_size)
        except asyncio.IncompleteReadError:
            await self._write_http_response(writer, status=400, body='incomplete body')
            return

        try:
            await self._handle_incoming_bytes(body)
        except Exception as exc:  # pragma: no cover
            logger.exception('Failed to process inbound IPC event: %s', exc)
            await self._write_http_response(writer, status=500, body='failed to process event')
            return

        await self._write_http_response(writer, status=202, body='accepted')

    async def _handle_incoming_bytes(self, payload: bytes) -> None:
        message = json.loads(payload.decode('utf-8'))
        event = BaseEvent[Any].model_validate(message).event_reset()
        self._inbound_bus.emit(event)

    async def _send_unix(self, endpoint: _Endpoint, payload: dict[str, Any]) -> None:
        socket_path = endpoint.path or ''
        if not socket_path:
            raise ValueError(f'Invalid unix endpoint: {endpoint.raw}')

        _reader, writer = await asyncio.open_unix_connection(path=socket_path)
        writer.write(json.dumps(payload, separators=(',', ':')).encode('utf-8'))
        writer.write(b'\n')
        await writer.drain()
        writer.close()
        await writer.wait_closed()

    async def _send_http(self, endpoint: _Endpoint, payload: dict[str, Any]) -> None:
        payload_bytes = json.dumps(payload, separators=(',', ':')).encode('utf-8')
        request = Request(
            endpoint.raw,
            data=payload_bytes,
            headers={
                'content-type': 'application/json',
                'content-length': str(len(payload_bytes)),
            },
            method='POST',
        )

        def _post() -> int:
            with urlopen(request, timeout=10) as response:
                return int(response.status)

        status_code = await asyncio.to_thread(_post)
        if status_code < 200 or status_code >= 300:
            raise RuntimeError(f'IPC HTTP send failed with status {status_code}: {endpoint.raw}')

    @staticmethod
    async def _write_http_response(writer: asyncio.StreamWriter, *, status: int, body: str) -> None:
        reasons = {
            202: 'Accepted',
            400: 'Bad Request',
            404: 'Not Found',
            405: 'Method Not Allowed',
            500: 'Internal Server Error',
        }
        reason = reasons.get(status, 'OK')
        body_bytes = body.encode('utf-8')
        headers = [
            f'HTTP/1.1 {status} {reason}',
            f'content-length: {len(body_bytes)}',
            'content-type: text/plain; charset=utf-8',
            'connection: close',
            '',
            '',
        ]
        writer.write('\r\n'.join(headers).encode('utf-8'))
        writer.write(body_bytes)
        await writer.drain()
        writer.close()
        await writer.wait_closed()


class HTTPEventBridge(EventBridge):
    """Bridge events over HTTP(S) endpoints."""

    def __init__(self, send_to: str | None = None, listen_on: str | None = None, *, name: str | None = None):
        if send_to and _parse_endpoint(send_to).scheme == 'unix':
            raise ValueError('HTTPEventBridge send_to must be http:// or https://')
        if listen_on and _parse_endpoint(listen_on).scheme != 'http':
            raise ValueError('HTTPEventBridge listen_on must be http://')
        super().__init__(send_to=send_to, listen_on=listen_on, name=name or f'HTTPEventBridge_{uuid7str()[-8:]}')


class SocketEventBridge(EventBridge):
    """Bridge events over a unix domain socket path."""

    def __init__(self, path: str | None = None, *, name: str | None = None):
        if path is None:
            send_to = None
            listen_on = None
        else:
            normalized = path[7:] if path.startswith('unix://') else path
            if not normalized:
                raise ValueError('SocketEventBridge path must not be empty')
            send_to = f'unix://{normalized}'
            listen_on = f'unix://{normalized}'

        super().__init__(send_to=send_to, listen_on=listen_on, name=name or f'SocketEventBridge_{uuid7str()[-8:]}')


def __getattr__(name: str) -> Any:
    module_name = _LAZY_BRIDGE_MODULES.get(name)
    if module_name is None:
        raise AttributeError(name)
    module = importlib.import_module(module_name, __package__)
    value = getattr(module, name)
    globals()[name] = value
    return value



import asyncio
import hashlib
import json
import logging
import os
import re
import tempfile
import threading
import time
from collections.abc import Callable, Coroutine
from functools import wraps
from pathlib import Path
from types import ModuleType
from typing import Any, Literal, ParamSpec, TypeVar, cast

psutil: ModuleType | None
try:
    import psutil as _psutil
except ImportError:
    psutil = None
else:
    psutil = _psutil

PSUTIL_AVAILABLE: bool = psutil is not None


logger = logging.getLogger(__name__)


T = TypeVar('T')
P = ParamSpec('P')
RetryErrorMatcher = type[Exception] | re.Pattern[str]
RetryOnErrors = list[RetryErrorMatcher] | tuple[RetryErrorMatcher, ...]

# Global semaphore registry for retry decorator
GLOBAL_RETRY_SEMAPHORES: dict[str, asyncio.Semaphore] = {}
GLOBAL_RETRY_SEMAPHORE_LOCK = threading.Lock()

# Multiprocess semaphore support
MULTIPROCESS_SEMAPHORE_DIR = Path(tempfile.gettempdir()) / 'browser_use_semaphores'
MULTIPROCESS_SEMAPHORE_DIR.mkdir(exist_ok=True)
MULTIPROCESS_STALE_LOCK_SECONDS = 300.0

# Global overload detection state
_last_overload_check = 0.0
_overload_check_interval = 5.0  # Check every 5 seconds
_active_retry_operations = 0
_active_operations_lock = threading.Lock()


def _check_system_overload() -> tuple[bool, str]:
    """Check if system is overloaded and return (is_overloaded, reason)"""
    if not PSUTIL_AVAILABLE:
        return False, ''

    assert psutil is not None
    try:
        # Get system stats
        cpu_percent = psutil.cpu_percent(interval=0.1)
        memory = psutil.virtual_memory()

        # Check thresholds
        reasons: list[str] = []
        is_overloaded = False

        if cpu_percent > 85:
            is_overloaded = True
            reasons.append(f'CPU: {cpu_percent:.1f}%')

        if memory.percent > 85:
            is_overloaded = True
            reasons.append(f'Memory: {memory.percent:.1f}%')

        # Check number of concurrent operations
        with _active_operations_lock:
            if _active_retry_operations > 30:
                is_overloaded = True
                reasons.append(f'Active operations: {_active_retry_operations}')

        return is_overloaded, ', '.join(reasons)
    except Exception:
        return False, ''


def _get_semaphore_key(
    base_name: str,
    semaphore_scope: Literal['multiprocess', 'global', 'class', 'instance'],
    args: tuple[Any, ...],
) -> str:
    """Determine the semaphore key based on scope."""
    if semaphore_scope == 'multiprocess':
        return base_name
    elif semaphore_scope == 'global':
        return base_name
    elif semaphore_scope == 'class' and args and hasattr(args[0], '__class__'):
        class_name = args[0].__class__.__name__
        return f'{class_name}.{base_name}'
    elif semaphore_scope == 'instance' and args:
        instance_id = id(args[0])
        return f'{instance_id}.{base_name}'
    else:
        # Fallback to global if we can't determine scope
        return base_name


def _get_or_create_semaphore(
    sem_key: str,
    semaphore_limit: int,
    semaphore_scope: Literal['multiprocess', 'global', 'class', 'instance'],
) -> Any:
    """Get or create a semaphore based on scope."""
    if semaphore_scope == 'multiprocess':
        return sem_key
    else:
        with GLOBAL_RETRY_SEMAPHORE_LOCK:
            if sem_key not in GLOBAL_RETRY_SEMAPHORES:
                GLOBAL_RETRY_SEMAPHORES[sem_key] = asyncio.Semaphore(semaphore_limit)
            return GLOBAL_RETRY_SEMAPHORES[sem_key]


def _calculate_semaphore_timeout(
    semaphore_timeout: float | None,
    timeout: float | None,
    semaphore_limit: int,
) -> float | None:
    """Calculate the timeout for semaphore acquisition."""
    if semaphore_timeout is not None:
        return semaphore_timeout
    if timeout is None:
        return None
    # Default aligns with TS: timeout * max(1, semaphore_limit - 1)
    return timeout * max(1, semaphore_limit - 1)


def _callable_name(func: Callable[..., Any]) -> str:
    """Return a stable name for logs even for callable instances."""
    return getattr(func, '__name__', func.__class__.__name__)


def _resolve_semaphore_name(
    func_name: str,
    semaphore_name: str | Callable[..., str] | None,
    args: tuple[Any, ...],
) -> str:
    """Resolve semaphore name from a static name or call-time getter."""
    if semaphore_name is None:
        return func_name
    if isinstance(semaphore_name, str):
        return semaphore_name
    return semaphore_name(*args)


def _matches_retry_on_error(error: Exception, retry_on_errors: RetryOnErrors | None) -> bool:
    """Return True when an error matches any configured retry matcher."""
    if not retry_on_errors:
        return True

    error_text = f'{error.__class__.__name__}: {error}'
    for matcher in retry_on_errors:
        if isinstance(matcher, re.Pattern):
            if matcher.search(error_text):
                return True
            continue
        if isinstance(error, matcher):
            return True

    return False


async def _acquire_multiprocess_semaphore(
    semaphore: Any,
    sem_timeout: float | None,
    sem_key: str,
    semaphore_lax: bool,
    semaphore_limit: int,
    timeout: float | None,
) -> tuple[bool, Any]:
    """Acquire a cross-process semaphore using shared slot files.

    Each slot is a file created with O_EXCL. The file content stores the owning
    pid and a per-acquisition token so release can verify ownership. This format
    is mirrored in abxbus-ts so the same semaphore name can contend across
    Python and JS processes.
    """
    del semaphore

    start_time = time.time()
    retry_delay = 0.1  # Start with 100ms
    backoff_factor = 2.0
    has_timeout = sem_timeout is not None and sem_timeout > 0
    lock_prefix = hashlib.sha256(sem_key.encode()).hexdigest()[:40]
    MULTIPROCESS_SEMAPHORE_DIR.mkdir(exist_ok=True, parents=True)

    while True:
        elapsed = time.time() - start_time
        remaining_time: float | None = (sem_timeout - elapsed) if has_timeout and sem_timeout is not None else None
        if remaining_time is not None and remaining_time <= 0:
            break

        for slot in range(semaphore_limit):
            slot_file = MULTIPROCESS_SEMAPHORE_DIR / f'{lock_prefix}.{slot:02d}.lock'
            token = f'{os.getpid()}:{time.time_ns()}:{threading.get_ident()}'
            owner = json.dumps(
                {
                    'token': token,
                    'pid': os.getpid(),
                    'semaphore_name': sem_key,
                    'created_at_ms': int(time.time() * 1000),
                }
            )

            try:
                fd = os.open(slot_file, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o600)
                with os.fdopen(fd, 'w', encoding='utf-8') as handle:
                    handle.write(owner)
                    handle.flush()
                    os.fsync(handle.fileno())
                return True, (slot_file, token)
            except FileExistsError:
                try:
                    raw = slot_file.read_text(encoding='utf-8').strip()
                    current_owner: dict[str, Any] | None = json.loads(raw) if raw else None
                    if not isinstance(current_owner, dict):
                        current_owner = None
                    current_pid = current_owner.get('pid') if current_owner else None
                    if isinstance(current_pid, int):
                        try:
                            os.kill(current_pid, 0)
                            continue
                        except OSError:
                            pass

                    slot_age = time.time() - slot_file.stat().st_mtime
                    if isinstance(current_pid, int) or slot_age >= MULTIPROCESS_STALE_LOCK_SECONDS:
                        slot_file.unlink(missing_ok=True)
                except Exception:
                    pass
            except FileNotFoundError:
                MULTIPROCESS_SEMAPHORE_DIR.mkdir(exist_ok=True, parents=True)
                continue

        sleep_for = retry_delay if remaining_time is None else min(retry_delay, remaining_time)
        if sleep_for > 0:
            await asyncio.sleep(sleep_for)
        retry_delay = min(retry_delay * backoff_factor, 1.0)

    # Timeout reached
    if not semaphore_lax:
        timeout_str = f', timeout={timeout}s per operation' if timeout is not None else ''
        raise TimeoutError(
            f'Failed to acquire multiprocess semaphore "{sem_key}" within {sem_timeout}s (limit={semaphore_limit}{timeout_str})'
        )
    logger.warning(
        f'Failed to acquire multiprocess semaphore "{sem_key}" after {sem_timeout:.1f}s, proceeding without concurrency limit'
    )
    return False, None


async def _acquire_asyncio_semaphore(
    semaphore: asyncio.Semaphore,
    sem_timeout: float | None,
    sem_key: str,
    semaphore_lax: bool,
    semaphore_limit: int,
    timeout: float | None,
    sem_start: float,
) -> bool:
    """Acquire an asyncio semaphore."""
    if sem_timeout is None or sem_timeout <= 0:
        await semaphore.acquire()
        return True

    try:
        async with asyncio.timeout(sem_timeout):
            await semaphore.acquire()
            return True
    except TimeoutError:
        sem_wait_time = time.time() - sem_start
        if not semaphore_lax:
            timeout_str = f', timeout={timeout}s per operation' if timeout is not None else ''
            raise TimeoutError(
                f'Failed to acquire semaphore "{sem_key}" within {sem_timeout}s (limit={semaphore_limit}{timeout_str})'
            )
        logger.warning(
            f'Failed to acquire semaphore "{sem_key}" after {sem_wait_time:.1f}s, proceeding without concurrency limit'
        )
        return False


async def _execute_with_retries(
    func: Callable[P, Coroutine[Any, Any, T]],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    max_attempts: int,
    timeout: float | None,
    retry_after: float,
    retry_backoff_factor: float,
    retry_on_errors: RetryOnErrors | None,
    start_time: float,
    sem_start: float,
    semaphore_limit: int | None,
) -> T:
    """Execute the function with retry logic."""
    func_name = _callable_name(func)
    func_runner = cast(Callable[..., Coroutine[Any, Any, T]], func)
    for attempt in range(1, max_attempts + 1):
        try:
            # Execute with per-attempt timeout
            if timeout is not None and timeout > 0:
                async with asyncio.timeout(timeout):
                    return await func_runner(*args, **kwargs)
            return await func_runner(*args, **kwargs)

        except Exception as e:
            # Check if we should retry this exception
            if not _matches_retry_on_error(e, retry_on_errors):
                raise

            if attempt < max_attempts:
                # Calculate wait time with backoff
                current_wait = retry_after * (retry_backoff_factor ** (attempt - 1))

                # Only log warning on the final retry attempt (second-to-last overall attempt)
                if attempt == max_attempts - 1:
                    logger.warning(
                        f'{func_name} failed (attempt {attempt}/{max_attempts}): '
                        f'{type(e).__name__}: {e}. Waiting {current_wait:.1f}s before retry...'
                    )
                if current_wait > 0:
                    await asyncio.sleep(current_wait)
            else:
                # Final failure
                total_time = time.time() - start_time
                sem_wait = time.time() - sem_start - total_time if semaphore_limit else 0
                sem_str = f'Semaphore wait: {sem_wait:.1f}s. ' if sem_wait > 0 else ''
                logger.error(
                    f'{func_name} failed after {max_attempts} attempts over {total_time:.1f}s. '
                    f'{sem_str}Final error: {type(e).__name__}: {e}'
                )
                raise

    # This should never be reached, but satisfies type checker
    raise RuntimeError('Unexpected state in retry logic')


def _track_active_operations(increment: bool = True) -> None:
    """Track active retry operations."""
    global _active_retry_operations
    with _active_operations_lock:
        if increment:
            _active_retry_operations += 1
        else:
            _active_retry_operations = max(0, _active_retry_operations - 1)


def _check_system_overload_if_needed() -> None:
    """Check for system overload if enough time has passed since last check."""
    global _last_overload_check
    current_time = time.time()
    if current_time - _last_overload_check > _overload_check_interval:
        _last_overload_check = current_time
        is_overloaded, reason = _check_system_overload()
        if is_overloaded:
            logger.warning(f'⚠️  System overload detected: {reason}. Consider reducing concurrent operations to prevent hanging.')


def retry(
    retry_after: float = 0,
    max_attempts: int = 1,
    timeout: float | None = None,
    retry_on_errors: RetryOnErrors | None = None,
    retry_backoff_factor: float = 1.0,
    semaphore_limit: int | None = None,
    semaphore_name: str | Callable[..., str] | None = None,
    semaphore_lax: bool = True,
    semaphore_scope: Literal['multiprocess', 'global', 'class', 'instance'] = 'global',
    semaphore_timeout: float | None = None,
):
    """
        Retry decorator with semaphore support for async functions.

        Args:
                retry_after: Seconds to wait between retries
                max_attempts: Total attempts including the initial call (1 = no retries)
                timeout: Per-attempt timeout in seconds (`None` = no per-attempt timeout)
                retry_on_errors: Error matchers to retry on (Exception subclasses or compiled regexes)
                retry_backoff_factor: Multiplier for retry delay after each attempt (1.0 = no backoff)
                semaphore_limit: Max concurrent executions (creates semaphore if needed)
                semaphore_name: Name for semaphore (defaults to function name), or callable receiving function args
                semaphore_lax: If True, continue without semaphore on acquisition failure
                semaphore_scope: Scope for semaphore sharing:
                        - 'global': All calls share one semaphore (default)
                        - 'class': All instances of a class share one semaphore
                        - 'instance': Each instance gets its own semaphore
                        - 'multiprocess': All processes on the machine share one semaphore
                semaphore_timeout: Max time to wait for semaphore acquisition
                                   (`None` => `timeout * max(1, limit - 1)` when timeout is set, else unbounded)

        Example:
                @retry(retry_after=3, max_attempts=3, timeout=5, semaphore_limit=3, semaphore_scope='instance')
                async def some_function(self, ...):
                        # Limited to 5s per attempt, up to 3 total attempts
                        # Max 3 concurrent executions per instance

    Notes:
                - semaphore acquisition happens once at start time, it is not retried
                - semaphore_timeout is only used if semaphore_limit is set.
                - if semaphore_timeout is set to 0, it waits forever for a semaphore slot.
                - if semaphore_timeout is None and timeout is None, semaphore acquisition wait is unbounded.
    """

    def decorator(func: Callable[P, Coroutine[Any, Any, T]]) -> Callable[P, Coroutine[Any, Any, T]]:
        func_name = _callable_name(func)
        effective_max_attempts = max(1, max_attempts)
        effective_retry_after = max(0, retry_after)
        effective_semaphore_limit = semaphore_limit if semaphore_limit is not None and semaphore_limit > 0 else None

        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            # Initialize semaphore-related variables
            semaphore: Any = None
            semaphore_acquired = False
            multiprocess_lock: Any = None
            sem_start = time.time()

            # Handle semaphore if specified
            if effective_semaphore_limit is not None:
                # Get semaphore key and create/retrieve semaphore
                base_name = _resolve_semaphore_name(func_name, semaphore_name, tuple(args))
                sem_key = _get_semaphore_key(base_name, semaphore_scope, tuple(args))
                semaphore = _get_or_create_semaphore(sem_key, effective_semaphore_limit, semaphore_scope)

                # Calculate timeout for semaphore acquisition
                sem_timeout = _calculate_semaphore_timeout(semaphore_timeout, timeout, effective_semaphore_limit)

                # Acquire semaphore based on type
                if semaphore_scope == 'multiprocess':
                    semaphore_acquired, multiprocess_lock = await _acquire_multiprocess_semaphore(
                        semaphore, sem_timeout, sem_key, semaphore_lax, effective_semaphore_limit, timeout
                    )
                else:
                    semaphore_acquired = await _acquire_asyncio_semaphore(
                        semaphore, sem_timeout, sem_key, semaphore_lax, effective_semaphore_limit, timeout, sem_start
                    )

            # Track active operations and check system overload
            _track_active_operations(increment=True)
            _check_system_overload_if_needed()

            # Execute function with retries
            start_time = time.time()
            try:
                return await _execute_with_retries(
                    func,
                    tuple(args),
                    dict(kwargs),
                    effective_max_attempts,
                    timeout,
                    effective_retry_after,
                    retry_backoff_factor,
                    retry_on_errors,
                    start_time,
                    sem_start,
                    effective_semaphore_limit,
                )
            finally:
                # Clean up: decrement active operations and release semaphore
                _track_active_operations(increment=False)

                if semaphore_acquired and semaphore:
                    try:
                        if semaphore_scope == 'multiprocess' and multiprocess_lock:
                            slot_file, token = cast(tuple[Path, str], multiprocess_lock)
                            raw = slot_file.read_text(encoding='utf-8').strip() if slot_file.exists() else ''
                            current_owner: dict[str, Any] | None = json.loads(raw) if raw else None
                            if not isinstance(current_owner, dict):
                                current_owner = None
                            if current_owner and current_owner.get('token') == token:
                                slot_file.unlink(missing_ok=True)
                        elif semaphore:
                            semaphore.release()
                    except (FileNotFoundError, OSError) as e:
                        # Handle case where lock file was removed during operation
                        if isinstance(e, FileNotFoundError) or 'No such file or directory' in str(e):
                            logger.warning(f'Semaphore lock file disappeared during release, ignoring: {e}')
                        else:
                            # Log other OS errors but don't raise - we already completed the operation
                            logger.error(f'Error releasing semaphore: {e}')

        return wrapper

    return decorator


__all__ = [
    'MULTIPROCESS_SEMAPHORE_DIR',
    'retry',
]



from __future__ import annotations

import logging
from collections.abc import Awaitable, Callable
from datetime import UTC, datetime, timedelta
from typing import Any, Generic, Literal, TypeVar, overload

from .base_event import BaseEvent, UUIDStr
from .helpers import monotonic_datetime

BaseEventT = TypeVar('BaseEventT', bound=BaseEvent[Any])
TExpectedEvent = TypeVar('TExpectedEvent', bound=BaseEvent[Any])
EventPatternType = str | Literal['*'] | type[BaseEvent[Any]]

logger = logging.getLogger('abxbus')


class EventHistory(dict[UUIDStr, BaseEventT], Generic[BaseEventT]):
    """Ordered event history map with query and trim helpers."""

    __slots__ = ('max_history_size', 'max_history_drop', '_warned_about_dropping_uncompleted_events')

    def __init__(self, max_history_size: int | None = 100, max_history_drop: bool = False):
        super().__init__()
        self.max_history_size = max_history_size
        self.max_history_drop = max_history_drop
        self._warned_about_dropping_uncompleted_events = False

    def add_event(self, event: BaseEventT) -> None:
        self[event.event_id] = event

    def get_event(self, event_id: str) -> BaseEventT | None:
        return self.get(event_id)

    def remove_event(self, event_id: str) -> bool:
        if event_id not in self:
            return False
        del self[event_id]
        return True

    def has_event(self, event_id: str) -> bool:
        return event_id in self

    @staticmethod
    def normalize_event_pattern(event_pattern: EventPatternType) -> str:
        if event_pattern == '*':
            return '*'
        if isinstance(event_pattern, str):
            return event_pattern
        event_type_field = event_pattern.model_fields.get('event_type')
        event_type_default = event_type_field.default if event_type_field is not None else None
        if isinstance(event_type_default, str) and event_type_default not in ('', 'UndefinedEvent'):
            return event_type_default
        return event_pattern.__name__

    def event_is_child_of(self, event: BaseEvent[Any], ancestor: BaseEvent[Any]) -> bool:
        current_id = event.event_parent_id
        visited: set[str] = set()

        while current_id and current_id not in visited:
            if current_id == ancestor.event_id:
                return True
            visited.add(current_id)
            parent = self.get(current_id)
            if parent is None:
                return False
            current_id = parent.event_parent_id

        return False

    @overload
    async def find(
        self,
        event_type: type[TExpectedEvent],
        where: None = None,
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        event_is_child_of: Callable[[BaseEvent[Any], BaseEvent[Any]], bool] | None = None,
        wait_for_future_match: Callable[
            [str, Callable[[BaseEvent[Any]], bool], bool | float],
            Awaitable[BaseEvent[Any] | None],
        ]
        | None = None,
        **event_fields: Any,
    ) -> TExpectedEvent | None: ...

    @overload
    async def find(
        self,
        event_type: type[TExpectedEvent],
        where: Callable[[TExpectedEvent], bool],
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        event_is_child_of: Callable[[BaseEvent[Any], BaseEvent[Any]], bool] | None = None,
        wait_for_future_match: Callable[
            [str, Callable[[BaseEvent[Any]], bool], bool | float],
            Awaitable[BaseEvent[Any] | None],
        ]
        | None = None,
        **event_fields: Any,
    ) -> TExpectedEvent | None: ...

    @overload
    async def find(
        self,
        event_type: str | Literal['*'],
        where: Callable[[BaseEvent[Any]], bool] | None = None,
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        event_is_child_of: Callable[[BaseEvent[Any], BaseEvent[Any]], bool] | None = None,
        wait_for_future_match: Callable[
            [str, Callable[[BaseEvent[Any]], bool], bool | float],
            Awaitable[BaseEvent[Any] | None],
        ]
        | None = None,
        **event_fields: Any,
    ) -> BaseEvent[Any] | None: ...

    async def find(
        self,
        event_type: EventPatternType,
        where: Callable[[Any], bool] | None = None,
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        event_is_child_of: Callable[[BaseEvent[Any], BaseEvent[Any]], bool] | None = None,
        wait_for_future_match: Callable[
            [str, Callable[[BaseEvent[Any]], bool], bool | float],
            Awaitable[BaseEvent[Any] | None],
        ]
        | None = None,
        **event_fields: Any,
    ) -> BaseEvent[Any] | None:
        resolved_past_input = True if past is None else past
        if isinstance(resolved_past_input, timedelta):
            resolved_past: bool | float = max(0.0, resolved_past_input.total_seconds())
        elif isinstance(resolved_past_input, bool):
            resolved_past = resolved_past_input
        else:
            resolved_past = max(0.0, float(resolved_past_input))

        resolved_future_input = False if future is None else future
        if isinstance(resolved_future_input, bool):
            resolved_future: bool | float = resolved_future_input
        else:
            resolved_future = max(0.0, float(resolved_future_input))

        if resolved_past is False and resolved_future is False:
            return None

        event_key = self.normalize_event_pattern(event_type)
        where_predicate: Callable[[BaseEvent[Any]], bool]
        if where is None:
            where_predicate = lambda _: True
        else:
            where_predicate = where

        child_check = event_is_child_of or self.event_is_child_of
        cutoff: str | None = None
        if resolved_past is not True:
            cutoff_dt = datetime.now(UTC) - timedelta(seconds=float(resolved_past))
            cutoff = monotonic_datetime(cutoff_dt.isoformat().replace('+00:00', 'Z'))
        missing = object()

        def matches(event: BaseEvent[Any]) -> bool:
            return (
                (event_key == '*' or event.event_type == event_key)
                and (child_of is None or child_check(event, child_of))
                and all(
                    getattr(event, field_name, missing) == expected_value for field_name, expected_value in event_fields.items()
                )
                and where_predicate(event)
            )

        if resolved_past is not False:
            events = list(self.values())
            for event in reversed(events):
                if cutoff is not None and event.event_created_at < cutoff:
                    continue
                if matches(event):
                    return event

        if resolved_future is False or wait_for_future_match is None:
            return None

        return await wait_for_future_match(event_key, matches, resolved_future)

    def trim_event_history(
        self,
        *,
        on_remove: Callable[[BaseEventT], None] | None = None,
        owner_label: str | None = None,
    ) -> int:
        if self.max_history_size is None:
            return 0

        if self.max_history_size == 0:
            completed_event_ids = [event_id for event_id, event in self.items() if event.event_status == 'completed']
            removed_count = 0
            for event_id in completed_event_ids:
                event = self.get(event_id)
                if event is None:
                    continue
                del self[event_id]
                if on_remove:
                    on_remove(event)
                removed_count += 1
            return removed_count

        if not self.max_history_drop or len(self) <= self.max_history_size:
            return 0

        remaining_overage = len(self) - self.max_history_size
        removed_count = 0

        def remove_event(event_id: str, event: BaseEventT) -> None:
            nonlocal removed_count
            del self[event_id]
            if on_remove:
                on_remove(event)
            removed_count += 1

        for event_id, event in list(self.items()):
            if remaining_overage <= 0:
                break
            if event.event_status != 'completed':
                continue
            remove_event(event_id, event)
            remaining_overage -= 1

        dropped_uncompleted = 0
        for event_id, event in list(self.items()):
            if remaining_overage <= 0:
                break
            if event.event_status != 'completed':
                dropped_uncompleted += 1
            remove_event(event_id, event)
            remaining_overage -= 1

        if dropped_uncompleted > 0 and not self._warned_about_dropping_uncompleted_events:
            self._warned_about_dropping_uncompleted_events = True
            owner = owner_label or 'EventBus'
            logger.warning(
                '[abxbus] ⚠️ Bus %s has exceeded max_history_size=%s and is dropping oldest history entries '
                '(even uncompleted events). Increase max_history_size or set max_history_drop=False to reject.',
                owner,
                self.max_history_size,
            )

        return removed_count



import asyncio
import contextvars
from contextlib import asynccontextmanager, contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Protocol, TypeVar

from abxbus.base_event import BaseEvent, EventConcurrencyMode, EventHandlerConcurrencyMode, EventResult

if TYPE_CHECKING:
    from abxbus.event_bus import EventBus

T_EventResultType = TypeVar('T_EventResultType')


class LockManagerProtocol(Protocol):
    """Minimal lock API required by EventBus runtime execution paths."""

    def get_lock_for_event(self, bus: 'EventBus', event: BaseEvent[T_EventResultType]) -> 'ReentrantLock | None':
        """Return the concrete event-level lock object or ``None`` for parallel mode."""
        ...

    def get_lock_for_event_handler(
        self,
        bus: 'EventBus',
        event: BaseEvent[T_EventResultType],
        event_result: EventResult[T_EventResultType],
    ) -> 'ReentrantLock | None':
        """Return the concrete handler-level lock object or ``None`` for parallel mode."""
        ...

    def _run_with_event_lock(self, bus: 'EventBus', event: BaseEvent[T_EventResultType]) -> Any:
        """Context manager for event-level lock scope."""
        ...

    def _run_with_handler_lock(
        self, bus: 'EventBus', event: BaseEvent[T_EventResultType], event_result: EventResult[T_EventResultType]
    ) -> Any:
        """Context manager for per-handler lock scope."""
        ...

    def _run_with_handler_dispatch_context(self, bus: 'EventBus', event: BaseEvent[T_EventResultType]) -> Any:
        """Context manager that mirrors held event-lock state into dispatch context."""
        ...


class ReentrantLock:
    """Context-aware re-entrant lock over an asyncio semaphore.

    Lifecycle:
    1. `__aenter__` acquires the semaphore when this context does not already hold
       the lock id.
    2. Nested entries in the same context only bump the local depth counter.
    3. `__aexit__` decrements depth and releases semaphore at depth zero.
    """

    # Context variable storing lock-id -> re-entrant depth for the current async context.
    _held_lock_depths: ContextVar[dict[int, int]] = ContextVar('held_lock_depths', default={})

    def __init__(self):
        self._semaphore: asyncio.Semaphore | None = None
        self._loop: asyncio.AbstractEventLoop | None = None
        self._lock_id = id(self)

    def _get_semaphore(self) -> asyncio.Semaphore:
        """Get or create the semaphore for the current event loop."""
        current_loop = asyncio.get_running_loop()
        if self._semaphore is None or self._loop != current_loop:
            # Create new semaphore for this event loop
            self._semaphore = asyncio.Semaphore(1)
            self._loop = current_loop
        return self._semaphore

    def _depth(self) -> int:
        return ReentrantLock._held_lock_depths.get().get(self._lock_id, 0)

    def _set_depth(self, depth: int) -> None:
        current = ReentrantLock._held_lock_depths.get()
        updated = dict(current)
        if depth <= 0:
            updated.pop(self._lock_id, None)
        else:
            updated[self._lock_id] = depth
        ReentrantLock._held_lock_depths.set(updated)

    def mark_held_in_current_context(self) -> contextvars.Token[dict[int, int]]:
        """Temporarily mark this lock as already held in the current context.

        Used when a handler runs in a copied dispatch context and needs re-entrant
        lock behavior to match the parent processing context.
        """
        current = ReentrantLock._held_lock_depths.get()
        updated = dict(current)
        updated[self._lock_id] = updated.get(self._lock_id, 0) + 1
        return ReentrantLock._held_lock_depths.set(updated)

    @staticmethod
    def reset_context_mark(token: contextvars.Token[dict[int, int]]) -> None:
        """Undo a prior `mark_held_in_current_context` update."""
        ReentrantLock._held_lock_depths.reset(token)

    async def __aenter__(self):
        depth = self._depth()
        if depth > 0:
            self._set_depth(depth + 1)
            return self

        # Acquire the lock
        await self._get_semaphore().acquire()
        self._set_depth(1)
        return self

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any) -> None:
        depth = self._depth()
        if depth <= 0:
            return

        next_depth = depth - 1
        self._set_depth(next_depth)
        if next_depth == 0:
            self._get_semaphore().release()

    def locked(self) -> bool:
        """Check if the lock is currently held."""
        # If semaphore doesn't exist yet or is from a different loop, it's not locked
        try:
            current_loop = asyncio.get_running_loop()
            if self._semaphore is None or self._loop != current_loop:
                return False
            return self._semaphore.locked()
        except RuntimeError:
            # No running loop, can't check
            return False


class LockManager:
    """Centralized lock/semaphore policy for event and handler execution.

    This manager owns lock resolution and all lock mutations. `EventBus` and
    handlers should use only these APIs instead of touching lock objects directly.
    """

    def get_lock_for_event(self, bus: 'EventBus', event: BaseEvent[T_EventResultType]) -> ReentrantLock | None:
        """Resolve the event-level lock for one event execution.

        Lifecycle:
        - Called before processing an event (runloop, step, queue-jump).
        - Returns `None` for `'parallel'`, so no lock is acquired.
        - Returns the shared class lock for `'global-serial'`.
        - Returns `bus.event_bus_serial_lock` for `'bus-serial'`.
        """
        resolved = event.event_concurrency or bus.event_concurrency
        if resolved == EventConcurrencyMode.PARALLEL:
            return None
        if resolved == EventConcurrencyMode.GLOBAL_SERIAL:
            return bus.event_global_serial_lock
        return bus.event_bus_serial_lock

    def get_lock_for_event_handler(
        self,
        bus: 'EventBus',
        event: BaseEvent[T_EventResultType],
        event_result: EventResult[T_EventResultType],
    ) -> ReentrantLock | None:
        """Resolve the per-event handler lock for one handler execution.

        Lifecycle:
        - Called inside `EventBus._run_handler` before running a handler.
        - Returns `None` for `'parallel'` handler mode.
        - Returns and lazily initializes the event handler lock for `'serial'`.
        """
        del event_result  # reserved for future mode-specific rules
        resolved = event.event_handler_concurrency or bus.event_handler_concurrency
        if resolved == EventHandlerConcurrencyMode.PARALLEL:
            return None
        current_lock = event._get_handler_lock()  # pyright: ignore[reportPrivateUsage]
        if current_lock is None:
            current_lock = ReentrantLock()
            event._set_handler_lock(current_lock)  # pyright: ignore[reportPrivateUsage]
        return current_lock

    @asynccontextmanager
    async def _run_with_event_lock(self, bus: 'EventBus', event: BaseEvent[T_EventResultType]):
        """Acquire/release the resolved event lock around event processing.

        Lifecycle:
        - Wraps event processing in runloop and manual `step()`.
        - No-op for `'parallel'` events.
        """
        lock = self.get_lock_for_event(bus, event)
        if lock is None:
            yield
            return
        async with lock:
            yield

    @asynccontextmanager
    async def _run_with_handler_lock(
        self, bus: 'EventBus', event: BaseEvent[T_EventResultType], event_result: EventResult[T_EventResultType]
    ):
        """Acquire/release the resolved per-event handler lock around one handler run.

        Lifecycle:
        - Used directly inside `EventResult.run_handler(...)`.
        - No-op for `'parallel'` handler mode.
        """
        lock = self.get_lock_for_event_handler(bus, event, event_result)
        if lock is None:
            yield
            return
        async with lock:
            yield

    @contextmanager
    def _run_with_handler_dispatch_context(self, bus: 'EventBus', event: BaseEvent[T_EventResultType]):
        """Mirror parent event-lock ownership into the current copied context.

        Lifecycle:
        - Used only by `EventResult.run_handler` when running handlers inside a copied
          dispatch context (`context=dispatch_context`).
        - Marks the resolved event lock as held in this context without acquiring
          the semaphore, enabling safe re-entry for awaited child events.
        """
        lock = self.get_lock_for_event(bus, event)
        if lock is None:
            yield
            return
        token = lock.mark_held_in_current_context()
        try:
            yield
        finally:
            ReentrantLock.reset_context_mark(token)



import asyncio
import inspect
import os
from collections.abc import Awaitable, Callable, Coroutine
from datetime import UTC, datetime
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, TypeAlias, cast, overload, runtime_checkable
from uuid import NAMESPACE_DNS, UUID, uuid5
from weakref import ref as weakref

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from typing_extensions import TypeVar

from abxbus.helpers import monotonic_datetime

if TYPE_CHECKING:
    from abxbus.base_event import BaseEvent


# TypeVar for BaseEvent and its subclasses
# We use contravariant=True because if a handler accepts BaseEvent,
# it can also handle any subclass of BaseEvent
T_Event = TypeVar('T_Event', bound='BaseEvent[Any]', contravariant=True, default='BaseEvent[Any]')

# For protocols with __func__ attributes, we need an invariant TypeVar
T_EventInvariant = TypeVar('T_EventInvariant', bound='BaseEvent[Any]', default='BaseEvent[Any]')
T_EventResult = TypeVar('T_EventResult', default=Any)
T_HandlerEvent = TypeVar('T_HandlerEvent', bound='BaseEvent[Any]', default='BaseEvent[Any]')
T_HandlerReturn = TypeVar('T_HandlerReturn', default=Any)


class EventHandlerCancelledError(asyncio.CancelledError):
    """Handler was cancelled before starting or before producing a result."""


class EventHandlerTimeoutError(TimeoutError):
    """Handler exceeded its configured handler timeout."""


class EventHandlerAbortedError(asyncio.CancelledError):
    """Handler was interrupted while running (for example by event hard-timeout)."""


class EventHandlerResultSchemaError(ValueError):
    """Handler returned a value incompatible with the event_result_type schema."""


@runtime_checkable
class EventHandlerFunc(Protocol[T_Event]):
    """Protocol for sync event handler functions."""

    def __call__(self, event: T_Event, /) -> Any: ...


@runtime_checkable
class AsyncEventHandlerFunc(Protocol[T_Event]):
    """Protocol for async event handler functions."""

    async def __call__(self, event: T_Event, /) -> Any: ...


@runtime_checkable
class EventHandlerMethod(Protocol[T_Event]):
    """Protocol for instance method event handlers."""

    def __call__(self, self_: Any, event: T_Event, /) -> Any: ...

    __self__: Any
    __name__: str


@runtime_checkable
class AsyncEventHandlerMethod(Protocol[T_Event]):
    """Protocol for async instance method event handlers."""

    async def __call__(self, self_: Any, event: T_Event, /) -> Any: ...

    __self__: Any
    __name__: str


@runtime_checkable
class EventHandlerClassMethod(Protocol[T_EventInvariant]):
    """Protocol for class method event handlers."""

    def __call__(self, cls: type[Any], event: T_EventInvariant, /) -> Any: ...

    __self__: type[Any]
    __name__: str
    __func__: Callable[[type[Any], T_EventInvariant], Any]


@runtime_checkable
class AsyncEventHandlerClassMethod(Protocol[T_EventInvariant]):
    """Protocol for async class method event handlers."""

    async def __call__(self, cls: type[Any], event: T_EventInvariant, /) -> Any: ...

    __self__: type[Any]
    __name__: str
    __func__: Callable[[type[Any], T_EventInvariant], Awaitable[Any]]


# Event handlers are normalized to bound single-argument callables at registration time.
EventHandlerCallable: TypeAlias = EventHandlerFunc['BaseEvent[Any]'] | AsyncEventHandlerFunc['BaseEvent[Any]']

# Normalized async callable shape used at call sites that require a Coroutine.
NormalizedEventHandlerCallable: TypeAlias = Callable[
    ['BaseEvent[T_EventResult]'],
    Coroutine[Any, Any, T_EventResult | 'BaseEvent[Any]' | None],
]

# Internal normalized one-argument callable used for invocation at runtime.
_InvokableEventHandlerCallable: TypeAlias = Callable[['BaseEvent[Any]'], Any | Awaitable[Any]]

# ContravariantEventHandlerCallable allows subtype-specific handlers.
ContravariantEventHandlerCallable: TypeAlias = EventHandlerFunc[T_Event] | AsyncEventHandlerFunc[T_Event]

HANDLER_ID_NAMESPACE: UUID = uuid5(NAMESPACE_DNS, 'abxbus-handler')


def _format_handler_source_path(path: str, line_no: int | None = None) -> str:
    normalized = str(Path(path).expanduser().resolve())
    home = str(Path.home())
    if normalized == home:
        display = '~'
    elif normalized.startswith(home + os.sep):
        display = f'~{normalized[len(home) :]}'
    else:
        display = normalized
    return f'{display}:{line_no}' if line_no else display


def _defined_class_method_name(target: Callable[..., Any]) -> str | None:
    qualname = getattr(target, '__qualname__', None)
    if not isinstance(qualname, str):
        return None

    parts = qualname.split('.')
    if len(parts) < 2 or parts[-2] == '<locals>':
        return None

    class_name, method_name = parts[-2], parts[-1]
    if class_name.isidentifier() and method_name.isidentifier():
        return f'{class_name}.{method_name}'
    return None


class _HandlerCacheKey:
    __slots__ = ('handler_ref', 'handler_id', '_hash')

    def __init__(self, handler: Callable[..., Any]) -> None:
        # Some callables override __eq__ without __hash__ and become unhashable.
        # Use identity-based hashing for a stable cache key without retaining handlers.
        self.handler_ref = weakref(handler)
        self.handler_id = id(handler)
        self._hash = self.handler_id

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _HandlerCacheKey):
            return False
        if self.handler_id != other.handler_id:
            return False
        return self.handler_ref() is other.handler_ref()


@lru_cache(maxsize=100)
def _get_callable_handler_file_path(handler_key: _HandlerCacheKey) -> str | None:
    """Best-effort, low-overhead source location for a handler callable."""
    handler = handler_key.handler_ref()
    if handler is None:
        return None
    target: Any = handler.__func__ if inspect.ismethod(handler) else handler
    target = inspect.unwrap(target)

    code_obj = getattr(target, '__code__', None)
    if code_obj is not None:
        file_path = getattr(code_obj, 'co_filename', None)
        line_no = getattr(code_obj, 'co_firstlineno', None)
        if isinstance(file_path, str) and file_path.strip():
            return _format_handler_source_path(file_path, int(line_no) if isinstance(line_no, int) else None)

    try:
        source_file = inspect.getsourcefile(target) or inspect.getfile(target)
    except (OSError, TypeError):
        source_file = None

    line_no: int | None = None
    try:
        _, line_no = inspect.getsourcelines(target)
    except (OSError, TypeError):
        line_no = None

    if isinstance(source_file, str) and source_file.strip():
        return _format_handler_source_path(source_file, line_no)

    module = inspect.getmodule(target)
    module_file = getattr(module, '__file__', None) if module is not None else None
    if isinstance(module_file, str) and module_file.strip():
        return _format_handler_source_path(module_file, line_no)

    return None


@overload
def _normalize_handler_callable(
    handler: Callable[[T_HandlerEvent], Coroutine[Any, Any, T_HandlerReturn]],
) -> Callable[[T_HandlerEvent], Coroutine[Any, Any, T_HandlerReturn]]: ...


@overload
def _normalize_handler_callable(
    handler: Callable[[T_HandlerEvent], Awaitable[T_HandlerReturn]],
) -> Callable[[T_HandlerEvent], Coroutine[Any, Any, T_HandlerReturn]]: ...


@overload
def _normalize_handler_callable(
    handler: Callable[[T_HandlerEvent], T_HandlerReturn],
) -> Callable[[T_HandlerEvent], Coroutine[Any, Any, T_HandlerReturn]]: ...


def _normalize_handler_callable(
    handler: Callable[[T_HandlerEvent], object],
) -> Callable[[T_HandlerEvent], Coroutine[Any, Any, T_HandlerReturn]]:
    """Normalize one handler callable to a single async call signature."""
    if not callable(handler):
        raise ValueError(f'Handler {handler!r} must be callable, got: {type(handler)}')

    if inspect.iscoroutinefunction(handler):
        return cast(Callable[[T_HandlerEvent], Coroutine[Any, Any, T_HandlerReturn]], handler)

    async def normalized_handler(event: T_HandlerEvent) -> T_HandlerReturn:
        handler_result = handler(event)
        # BaseEvent implements __await__ for ergonomic `await event`, but handler
        # return values of BaseEvent must be treated as plain results (forwarded
        # child event refs), not awaited here.
        if inspect.isawaitable(handler_result):
            from abxbus.base_event import BaseEvent

            if isinstance(handler_result, BaseEvent):
                return cast(T_HandlerReturn, handler_result)
            return cast(T_HandlerReturn, await handler_result)
        return cast(T_HandlerReturn, handler_result)

    return normalized_handler


class EventHandler(BaseModel):
    """Serializable metadata wrapper around a registered event handler callable."""

    model_config = ConfigDict(
        extra='forbid',
        arbitrary_types_allowed=True,
        validate_assignment=False,
        validate_default=True,
        revalidate_instances='always',
    )

    id: str = ''
    handler: EventHandlerCallable | None = Field(default=None, exclude=True, repr=False)
    handler_name: str = 'anonymous'
    handler_file_path: str | None = None
    handler_timeout: float | None = None
    handler_slow_timeout: float | None = None
    handler_registered_at: str = Field(default_factory=monotonic_datetime)
    event_pattern: str = '*'
    eventbus_name: str = 'EventBus'
    eventbus_id: str = '00000000-0000-0000-0000-000000000000'

    @field_validator('handler_name', mode='before')
    @classmethod
    def _validate_handler_name_field(cls, value: Any) -> str:
        if isinstance(value, str):
            normalized = value.strip()
            if normalized:
                return normalized
        return 'anonymous'

    @field_validator('eventbus_name')
    @classmethod
    def _validate_eventbus_name_field(cls, value: str) -> str:
        normalized = str(value)
        assert normalized.isidentifier() and not normalized.startswith('_'), f'Invalid event bus name: {value!r}'
        return normalized

    @field_validator('handler_registered_at', mode='before')
    @classmethod
    def _normalize_handler_registered_at(cls, value: Any) -> str:
        if isinstance(value, datetime):
            normalized_value = value.replace(tzinfo=UTC) if value.tzinfo is None else value.astimezone(UTC)
            normalized_input = normalized_value.isoformat().replace('+00:00', 'Z')
            return monotonic_datetime(normalized_input)
        if value is None:
            return monotonic_datetime()
        return monotonic_datetime(str(value))

    @property
    def eventbus_label(self) -> str:
        return f'{self.eventbus_name}#{self.eventbus_id[-4:]}'

    @staticmethod
    def get_callable_handler_name(handler: Callable[..., Any]) -> str:
        assert hasattr(handler, '__name__'), f'Handler {handler} has no __name__ attribute!'
        if inspect.ismethod(handler):
            defined_name = _defined_class_method_name(handler.__func__)
            if defined_name:
                return defined_name
            return f'{type(handler.__self__).__name__}.{handler.__name__}'
        if callable(handler):
            defined_name = _defined_class_method_name(handler)
            if defined_name:
                return defined_name
            handler_module = getattr(handler, '__module__', '<unknown>')
            handler_name = getattr(handler, '__name__', type(handler).__name__)
            return f'{handler_module}.{handler_name}'
        raise ValueError(f'Invalid handler: {handler} {type(handler)}, expected a function, coroutine, or method')

    @model_validator(mode='before')
    @classmethod
    def _populate_handler_name(cls, data: Any) -> Any:
        if not isinstance(data, dict):
            return data
        params = cast(dict[str, Any], data)
        if params.get('id') is None:
            params.pop('id', None)
        handler = params.get('handler')
        if handler is not None and not params.get('handler_name'):
            try:
                derived_name = cls.get_callable_handler_name(cast(EventHandlerCallable, handler))
                params['handler_name'] = derived_name.strip() or 'function'
            except Exception:
                params['handler_name'] = 'function'
        return params

    def model_post_init(self, __context: Any) -> None:
        if not self.id:
            self.id = self.compute_handler_id()

    @property
    def _handler_async(self) -> NormalizedEventHandlerCallable[Any] | None:
        """Return the normalized async callable view of `handler`."""
        if self.handler is None:
            return None
        return cast(
            NormalizedEventHandlerCallable[Any],
            _normalize_handler_callable(cast(Callable[[Any], object], self.handler)),
        )

    def compute_handler_id(self) -> str:
        """Match TS handler-id algorithm: uuidv5(seed, HANDLER_ID_NAMESPACE)."""
        file_path = self.handler_file_path or 'unknown'
        seed = f'{self.eventbus_id}|{self.handler_name}|{file_path}|{self.handler_registered_at}|{self.event_pattern}'
        return str(uuid5(HANDLER_ID_NAMESPACE, seed))

    @property
    def label(self) -> str:
        assert self.id, 'EventHandler.id must be set'
        return f'{self.handler_name}#{self.id[-4:]}'

    def __str__(self) -> str:
        has_name = self.handler_name and self.handler_name != 'anonymous'
        assert self.id, 'EventHandler.id must be set'
        display = f'{self.handler_name}()' if has_name else f'function#{self.id[-4:]}()'
        return f'{display} @ {self.handler_file_path}' if self.handler_file_path else display

    def __call__(self, event: 'BaseEvent[Any]') -> Any:
        if self.handler is None:
            raise RuntimeError(f'EventHandler {self.id} has no callable attached')
        handler_callable = cast(Callable[[Any], Any], self.handler)
        return handler_callable(event)

    @classmethod
    def from_callable(
        cls,
        *,
        handler: ContravariantEventHandlerCallable[Any],
        event_pattern: str,
        eventbus_name: str,
        eventbus_id: str,
        detect_handler_file_path: bool = True,
        id: str | None = None,
        handler_file_path: str | None = None,
        handler_timeout: float | None = None,
        handler_slow_timeout: float | None = None,
        handler_registered_at: str | datetime | None = None,
    ) -> 'EventHandler':
        resolved_file_path = handler_file_path
        if resolved_file_path is None and detect_handler_file_path:
            resolved_file_path = _get_callable_handler_file_path(_HandlerCacheKey(handler))

        handler_params: dict[str, Any] = {
            'handler': handler,
            'handler_file_path': resolved_file_path,
            'handler_registered_at': handler_registered_at,
            'event_pattern': event_pattern,
            'eventbus_name': eventbus_name,
            'eventbus_id': eventbus_id,
        }
        try:
            derived_name = cls.get_callable_handler_name(handler)
            handler_params['handler_name'] = derived_name.strip() or 'function'
        except Exception:
            handler_params['handler_name'] = 'function'
        if id is not None:
            handler_params['id'] = id
        if handler_timeout is not None:
            handler_params['handler_timeout'] = handler_timeout
        if handler_slow_timeout is not None:
            handler_params['handler_slow_timeout'] = handler_slow_timeout

        entry = cls(**handler_params)
        if not entry.id:
            entry.id = entry.compute_handler_id()
        return entry


__all__ = [
    'AsyncEventHandlerClassMethod',
    'AsyncEventHandlerFunc',
    'AsyncEventHandlerMethod',
    'ContravariantEventHandlerCallable',
    'EventHandler',
    'EventHandlerCallable',
    'EventHandlerClassMethod',
    'EventHandlerFunc',
    'EventHandlerMethod',
    'NormalizedEventHandlerCallable',
]
